#include <esp_now.h>
#include <WiFi.h>
#include <Arduino.h>
#include <ArduinoJson.h>
#include "SPIFFS.h"
#include <esp_wifi.h>
#include <WebServer.h>
#include <vector>
#include "ESPAsyncWebServer.h"
#include <math.h>

//Defines WiFi requirements
#define MAC_ADDRESS "58:CF:79:16:85:CC"
#define SERVER_PORT 80

//Define sizes for arrays
#define SIZE_DATA 15
#define SEND_SIZE_DATA 12
#define ARRAY_LENGTH 5

//Define types for SendMessage
#define FORCE_TRUE 0
#define GYRO_TRUE 1

using namespace std;
 
//Data structure, must match the sender structure
typedef struct struct_message {
  int number;
  double force[SIZE_DATA];
  double gyro[SEND_SIZE_DATA];
} struct_message;

const char* ssid = "SDNet";
const char* password = "CapstoneProject";

//Data structur for website
double maxForceValues[ARRAY_LENGTH];
double maxGyroValues[ARRAY_LENGTH];
double numberValues[ARRAY_LENGTH];
String severityValues[ARRAY_LENGTH];
String severity = "Default";

//WebServer server(SERVER_PORT);                
AsyncWebServer server(SERVER_PORT);

// Create a struct_message called myData
struct_message myData;

//Finds the absolute max value in the data
double findMax(double arr[], int size, bool type) {
  double max = abs(arr[0]); // Assume first element is the maximum
  for (int i = 1; i < size; i++) { // Loop through the remaining elements
    if (abs(arr[i]) > max) { // If an element is greater than the current max
      max = abs(arr[i]); // Store the new maximum
    }
  }
  if (type == FORCE_TRUE)
  {
    if (max > 65)
    {
      severity = "High";
    }
    else if ((max >= 45) && (max <= 65))
    {
      severity = "Medium";
    }
    else
    {
      severity = "Low";
    }
  }
  if (type == GYRO_TRUE)
  {
    if (max > 42)
    {
      severity = "High";
    }
    else if ((max >= 26) && (max <= 42))
    {
      severity = "Medium";
    }
    else
    {
      severity = "Low";
    }
  }

  return max; // Return the maximum value
}

//Callback function that will be executed when data is received
//Calls findmax() to find maximum value and define severity
//Moves data values into maxforceValues, etc to display on the website continously (look at getJSON)
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  memcpy(&myData, incomingData, sizeof(myData));
  Serial.print("Bytes received: ");
  Serial.println(len);

  Serial.print("Vector: ");
  for (int i = 0; i < 5; i++) {
    Serial.printf("%f ", myData.force[i]);
  }

  Serial.printf("Max: %f ", findMax(myData.force, SIZE_DATA, FORCE_TRUE));
  Serial.printf("Number: %d", myData.number);
  Serial.println();

  //Moves data values into website arrays
  for (int i = ARRAY_LENGTH - 1; i >= 1; i--) {
          maxForceValues[i] = maxForceValues[i-1];
          maxGyroValues[i] = maxGyroValues[i-1];
          numberValues[i] = numberValues[i-1];
          severityValues[i] = severityValues[i-1];
    }
    
    numberValues[0]=myData.number;
    maxForceValues[0]=findMax(myData.force, SIZE_DATA, FORCE_TRUE);
    maxGyroValues[0]=findMax(myData.gyro, SEND_SIZE_DATA, GYRO_TRUE);
    severityValues[0]=severity;

  //event.send ...
}

//Ensures that the Hub and Nodes are on the same WiFi channel
int32_t getWiFiChannel() {
  if (int32_t n = WiFi.scanNetworks()) {
      for (uint8_t i=0; i<n; i++) {
          if (!strcmp(ssid, WiFi.SSID(i).c_str())) {
              return WiFi.channel(i);
          }
      }
  }
  return 0;
}

//Updates the values displayed on the webpage
String getJSON(){
  String json = "{";
    json += "\"numberA\":" + String(numberValues[0])  + ",";
    json += "\"numberB\":" + String(numberValues[1])  + ",";
    json += "\"numberC\":" + String(numberValues[2])  + ",";
    json += "\"numberD\":" + String(numberValues[3])  + ",";
    json += "\"numberE\":" + String(numberValues[4])  + ",";
    json += "\"severityA\":\"" + severityValues[0]  + "\",";
    json += "\"severityB\":\"" + severityValues[1]  + "\",";
    json += "\"severityC\":\"" + severityValues[2]  + "\",";
     json += "\"severityD\":\"" + severityValues[3]  + "\",";
    json += "\"severityE\":\"" + severityValues[4]  + "\",";
    json += "\"maxForceA\":" + String(maxForceValues[0]) + ",";
    json += "\"maxForceB\":" + String(maxForceValues[1]) + ",";
    json += "\"maxForceC\":" + String(maxForceValues[2]) + ",";
    json += "\"maxForceD\":" + String(maxForceValues[3]) + ",";
    json += "\"maxForceE\":" + String(maxForceValues[4]) + ",";
    json += "\"maxGyroA\":" + String(maxGyroValues[0]) + ",";
    json += "\"maxGyroB\":" + String(maxGyroValues[1]) + ",";
    json += "\"maxGyroC\":" + String(maxGyroValues[2]) + ",";
    json += "\"maxGyroD\":" + String(maxGyroValues[3]) + ",";
    json += "\"maxGyroE\":" + String(maxGyroValues[4]);
    json += "}";
    return json;
    //myData.e.clear();
}

void setup() {
  // Initialize Serial Monitor
  Serial.begin(115200);

  // Set device as a Wi-Fi Station
  WiFi.mode(WIFI_AP_STA);
  Serial.println("Mac Address: ");
  Serial.println(WiFi.macAddress());

  // Connect to Wi-Fi and Display IP Address for Website
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to Wi-Fi...");
  }
  Serial.print("Connected. IP: ");
  Serial.println(WiFi.localIP());
  Serial.print("Wi-Fi Channel: ");
  Serial.println(WiFi.channel());

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/index.html", "text/html");
  });

  server.on("/data", HTTP_GET, [](AsyncWebServerRequest *request){
    String jsonData = getJSON();
    request->send(200, "application/json", jsonData);
  });  

  // Initializing ESP-NOW
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }
  
  // Once ESPNow is successfully Init, we will register for recv CB to get packer info
  esp_now_register_recv_cb(OnDataRecv);

  // Initialize and mount SPIFFS (allows us to upload HTML file)
  if (!SPIFFS.begin(true)) {
    Serial.println("An error has occurred while mounting SPIFFS");
    return;
  }

  // Start the server for the website
  server.begin();
}
 
void loop() {
  //Empty on purpose
}